/*<---------------------------------------------------------------------->*/
#include <windows.h>
#include <windowsx.h>
#include <commctrl.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <time.h>
#include "chrtbl_util.h"
#include "chrtbl_option.h"
#include "chrtbl_err.h"
/*<---------------------------------------------------------------------->*/
/* chr SJIS2byte1byteڂ̒lȂ1, łȂ0 */
static int is_sjis_2byte_1st(int chr)
{
	return (chr >= 0x81 && chr <= 0x9F) || (chr >= 0xE0 && chr <= 0xFC);
}
/*<---------------------------------------------------------------------->*/
/* chr SJIS2byte2byteڂ̒lȂ1, łȂ0 */
static int is_sjis_2byte_2nd(int chr)
{
	return (chr >= 0x40 && chr <= 0x7E) || (chr >= 0x80 && chr <= 0xFC);
}
/*<---------------------------------------------------------------------->*/
/* chr SJIS̔pJi̒lȂ1, łȂ0 */
static int is_sjis_hankakukana(int chr)
{
	return (chr >= 0xA1 && chr <= 0xDF);
}
/*<---------------------------------------------------------------------->*/
/*
 * ォEɊiqɕׂĕ`B
 * width	F`Ώۂ̉B̉ւ̕т̕ɒBƐ܂ԂB
 * chrs		FR[h̔zBSJIS1byte܂2byte̒lB
 * chrs_countFchrs ̗vfB
 * chr_w	Fiq̉BXW̑B
 * chr_h	Fiq̏cBYW̑B
 */
static void DrawMyFont(HDC hdc, int width, int *chrs, int chr_count, int chr_w, int chr_h)
{
	char str[3];
	int i,chr,len,x,y;

	x = y = 0;
	for(i=0;i<chr_count;i++){
		/* R[hP̕𐶐B */
		chr = chrs[i];
		/* ̔͂B */
		if(chr <= 0x7F){
			str[0] = (char)chr;
			str[1] = '\0';
			len = 1;
		}
		else{
			str[0] = (char)((chr>>8) & 0xff);
			str[1] = (char)(chr & 0xff);
			str[2] = '\0';
			len = 2;
		}
		TextOut(hdc, x, y, (LPCTSTR)str, len);
		x += chr_w;
		if(x+chr_w > width){
			x = 0;
			y += chr_h;
		}
	}
}
/*<---------------------------------------------------------------------->*/
/*
 * fp 當ǂݍ݁AoSJIS̕\\ɂ table ̗vfɃtO𗧂ĂB
 * table ̑傫 65536 ƂBȏ㖳ʂɂȂ̂Ŗܑ̖B
 *
 * ӁF
 * 2byte1byteځ2byte1byteځ2byte2byteڂƂтǂݍ܂
 * ܂ƂɓȂ낤B]ăoCỉɕĂ悤
 * t@C͈ȂB
 */
static void AppendSJISChrTable(FILE *fp, int *table)
{
	int chr,chr2,index;

	while((chr = fgetc(fp)) != EOF){
		/* SJIS2byte1byteڂH */
		if(is_sjis_2byte_1st(chr)){
			/* 2byteڂǂݍށB */
			chr2 = fgetc(fp);
			/* 2byteڂǂ߂ȂꍇAIB */
			if(chr2 == EOF)
				break;
			/* ꉞASJIS2byteڔB */
			if(is_sjis_2byte_2nd(chr2)){
				index = (chr<<8) + chr2;
				table[index] = 1;
			}
		}
		else if((chr >= 0x20 && chr <= 0x7E) || is_sjis_hankakukana(chr)){
			table[chr] = 1;
		}
	}
}
/*<---------------------------------------------------------------------->*/
/*
 * table ̔łvf̃CfbNX array Ɋi[B
 * table_size  table ̗vfB
 * array_size  array Ɋi[łvfB
 * ߂l array Ɋi[vfBG[-1B
 * ex. table = {0,1,0,1,1} Ȃ array = {1,3,4} ƂȂB
 */
static int MakeSJISChrArray(int *table, int table_size, int *array, int array_size)
{
	int i,count=0;

	for(i=0;i<table_size;i++){
		if(table[i]){
			if(count >= array_size)
				return -1;
			array[count] = i;
			count++;
		}
	}
	return count;
}
/*<---------------------------------------------------------------------->*/
/*
 * z̒lsŋ؂ăt@CɏށB
 */
static int SaveTableArray_Text(char *file, int *array, int array_size)
{
	FILE *fp;
	int i;

	fp = fopen(file, "w");
	if(!fp){
		print_err("Failed to open %s for output. (text mode)", file);
		return -1;
	}
	for(i=0;i<array_size;i++){
		fprintf(fp, "%x\n", array[i]);
	}
	fclose(fp);
	return 0;
}
/*<---------------------------------------------------------------------->*/
/*
 * SaveTableArray_Text() ̃oCiŁBint^̉16bit̒l𐶂ŏށB
 */
static int SaveTableArray_Binary(char *file, int *array, int array_size)
{
	FILE *fp;
	int i;
	unsigned char c;

	fp = fopen(file, "wb");
	if(!fp){
		print_err("Failed to open %s for binary output. (binary mode)", file);
		return -1;
	}
	for(i=0;i<array_size;i++){
		c = (unsigned char)((array[i]>>8) & 0xff);
		fwrite(&c, sizeof(char), 1, fp);
		c = (unsigned char)(array[i] & 0xff);
		fwrite(&c, sizeof(char), 1, fp);
	}
	fclose(fp);
	return 0;
}
/*<---------------------------------------------------------------------->*/
static BOOL SaveFile(LPCSTR szFileName, LPBITMAPINFOHEADER pbih, LPBYTE pBits)
{
	//save file
	HANDLE hf;
	BITMAPFILEHEADER hdr; 
	DWORD dwTmp;

	hf = CreateFile(szFileName, GENERIC_READ | GENERIC_WRITE, (DWORD)0, 
		NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, (HANDLE)NULL); 
	if (hf == INVALID_HANDLE_VALUE)
	{
		print_err("CeateFile Error\n");
		return FALSE;
	}

	hdr.bfType = 0x4d42;		// 0x42 = "B" 0x4d = "M" 
	// Compute the size of the entire file. 
	hdr.bfSize = (DWORD) (sizeof(BITMAPFILEHEADER) + pbih->biSize + pbih->biClrUsed * sizeof(RGBQUAD) + pbih->biSizeImage); 
	hdr.bfReserved1 = 0;
	hdr.bfReserved2 = 0;
	
	// Compute the offset to the array of color indices. 
	hdr.bfOffBits = (DWORD) sizeof(BITMAPFILEHEADER) + pbih->biSize + pbih->biClrUsed * sizeof (RGBQUAD);
	
	// Copy the BITMAPFILEHEADER into the .BMP file. 
	if (!WriteFile(hf, (LPVOID) &hdr, sizeof(BITMAPFILEHEADER), &dwTmp,  NULL))
	{
		print_err("WriteFile Error\n");
		return FALSE;
	}
	
	// Copy the BITMAPINFOHEADER and RGBQUAD array into the file. 
	if (!WriteFile(hf, (LPVOID)pbih, sizeof(BITMAPINFOHEADER) + pbih->biClrUsed * sizeof(RGBQUAD), (LPDWORD)&dwTmp, NULL))
	{
		print_err("WriteFile Error\n");
		return FALSE;
	}

	// Copy the array of color indices into the .BMP file. 
	DWORD cb = pbih->biSizeImage;
	if (!WriteFile(hf, pBits, (int)cb, &dwTmp, NULL))
	{
		print_err("WriteFile Error\n");
		return FALSE;
	}
	
	// Close the .BMP file. 
	if (!CloseHandle(hf)) 
	{
		print_err("CloseHandle Error\n");
		return FALSE;
	}

	return TRUE;
}
/*<---------------------------------------------------------------------->*/
BOOL GenerateChrBitmap(HWND hWnd, chrtbl_option_t *option)
{
	int chr_table[65536];
	int chr_array[32768];
	int chr_count;
	int i, j, chr_xcount, chr_ycount;
	FILE *fp;

	memset(chr_table,0,sizeof(chr_table));
	/* eLXgt@CǂŁAõtO𗧂ĂB */
	for(i=0; i<option->textfile_count; i++){
		fp = fopen(option->textfile[i], "r");
		if(!fp){
			print_err("textfile %s not found.", option->textfile[i]);
			return FALSE;
		}
		AppendSJISChrTable(fp,chr_table);
		fclose(fp);
	}
	/* eLXgt@C̎w肪ꍇASĂ̕\\̃tO𗧂ĂB */
	if(option->textfile_count == 0){
		/* pȂ */
		for(i=0x20;i<=0x7E;i++)
			chr_table[i] = 1;
		/* pJi */
		for(i=0xA1;i<=0xDF;i++)
			chr_table[i] = 1;
		/* ʓ|Ȃ̂ŕ̖So */
		for(i=0x81;i<=0x9F;i++){
			for(j=0x40;j<=0x7E;j++)
				chr_table[(i<<8)+j] = 1;
			for(j=0x80;j<=0xFC;j++)
				chr_table[(i<<8)+j] = 1;
		}
		for(i=0xE0;i<=0xFC;i++){
			for(j=0x40;j<=0x7E;j++)
				chr_table[(i<<8)+j] = 1;
			for(j=0x80;j<=0xFC;j++)
				chr_table[(i<<8)+j] = 1;
		}
	}
	chr_count = MakeSJISChrArray(chr_table,65536,chr_array,32768);

	/* w肳ꂽ摜cł͑ȂꍇA␳B */
	chr_xcount = option->width / option->char_width;
	chr_ycount = (chr_count + chr_xcount -1)/chr_xcount;
	if(option->height < option->char_height*chr_ycount){
		option->height = option->char_height*chr_ycount;
	}

	HDC hdc;
	hdc = CreateCompatibleDC(GetDC(hWnd));

	BITMAPINFO bi;

	/* ۂ̂Ƃ24bppȊO܂ȂB */
	bi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	bi.bmiHeader.biWidth = option->width;
	bi.bmiHeader.biHeight = option->height;
	bi.bmiHeader.biPlanes = 1;
	bi.bmiHeader.biBitCount = 24;
	bi.bmiHeader.biCompression = BI_RGB;

	WORD cClrBits = (WORD)(bi.bmiHeader.biPlanes * bi.bmiHeader.biBitCount); 
	if (cClrBits == 1) 
		cClrBits = 1; 
	else if (cClrBits <= 4) 
		cClrBits = 4; 
	else if (cClrBits <= 8) 
		cClrBits = 8; 
	else if (cClrBits <= 16) 
		cClrBits = 16; 
	else if (cClrBits <= 24) 
		cClrBits = 24; 
	else
		cClrBits = 32;
	bi.bmiHeader.biSizeImage = (bi.bmiHeader.biWidth + 7) / 8 
		* bi.bmiHeader.biHeight * cClrBits;

	bi.bmiHeader.biXPelsPerMeter = 0;
	bi.bmiHeader.biYPelsPerMeter = 0;
	bi.bmiHeader.biClrUsed = 0;
	bi.bmiHeader.biClrImportant = 0;

	BYTE* pBits;
	HBITMAP hbmp = CreateDIBSection(hdc, &bi, DIB_RGB_COLORS, (void**)&pBits, NULL, 0);

	SelectObject(hdc, hbmp);

	HFONT font;
	font = CreateFont(option->fontsize, 0, 0, 0, FW_NORMAL,
		FALSE, FALSE, FALSE, DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
		DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE, "lr SVbN");

	SelectObject(hdc, font);

	//draw fonts
	HGDIOBJ bg,hOldBrush;
	bg = GetStockObject(BLACK_BRUSH);
	hOldBrush = SelectObject(hdc, bg);
	Rectangle(hdc, 0, 0, option->width, option->height);
	SetTextColor(hdc, RGB(255, 255, 255));
	SetBkMode(hdc, TRANSPARENT);
	SetMapMode(hdc, MM_TEXT);
	SelectObject(hdc, hOldBrush);

	DrawMyFont(hdc, option->width, chr_array, chr_count, option->char_width, option->char_height);

	if(SaveFile(option->imagefile, &bi.bmiHeader, pBits) == FALSE)
	{
		DeleteDC(hdc);
		DeleteObject(font);
		DeleteObject(hbmp);
		return FALSE;
	}

	DeleteDC(hdc);
	DeleteObject(font);
	DeleteObject(hbmp);

	/* R[hz̒lXgt@CɏoB */
	if(option->charlistfile_text[0] != '\0'){
		SaveTableArray_Text(option->charlistfile_text, chr_array, chr_count);
	}
	if(option->charlistfile_binary[0] != '\0'){
		SaveTableArray_Binary(option->charlistfile_binary, chr_array, chr_count);
	}

	return TRUE;
}
/*<---------------------------------------------------------------------->*/
